/*----------------- IMA 4 PROJET P4: Train de vehicules ----------------*/
/*
Programme de contrôle d'un robot suiveur

Avancement:
- Entrées/Sorties 											FAIT / OK
- PWM 														FAIT / OK
- ADC (Voir si nécessaire)									FAIT / OK
- Interruption bouton / Affichage ID 						FAIT / OK
- Mesure distance Ultrason 									FAIT / OK
- Mesure IR / détermination position robot précédent 		FAIT / OK
- Asservissement 											FAIT / OK

*/


#include <avr/io.h>			
#include <avr/interrupt.h> 	
#include <util/delay.h>  	
#include <stdint.h>
#include <stdbool.h>
#define PRESCALER 		1024
#define TIME_SLOT	
#define NB_TICK			25 		//	timer 100us
#define BAUDRATE		103 	//	9600 bauds
#define	IRperiod0		130		//	Réservée au robot de tete
#define IRperiod1		210	
#define IRperiod2		330	
#define IRperiod3		410	
#define IRperiod4		520
#define CmdDist			20
#define Kp				20
#define gauche			0
#define milieu			1
#define droit			2
#define halfSpeed		30
#define fullSpeed		70


/* Variables globales */

//ID du robot
uint8_t ID = 1;

//Periodes Infrarouge
uint16_t IRperiodDetect = IRperiod0;
uint16_t IRperiodEmmit = IRperiod1;	
//Variables Infrarouge	
int16_t  IRcount[3];
uint8_t IRprec[3] = {255};
uint8_t IR[3];
bool IRcounting[3] = {false};
bool IRdetectedOnce[3] = {false};
bool IRdetected[3] = {false};
bool IRcollision[3] =  {false};
int16_t IRcountEmmit=0;
//Conversion Analogique/numérique
int8_t ADCchannel=0;
//Variables Ultrason
int16_t UScount=0;
uint8_t USdist=20;
//Dernière position du robot à suivre
uint8_t lastPos=gauche;



//PWM 8kHz sur pins 3(PD3/OC2B) et 11(PB3/OC2A) pour commande moteurs
void init_pwm()
{	
	//PD3 et PB3 en sortie
	DDRD|=0x08;
	DDRB|=0x08;

	//Registres de comparaison à 0 
	OCR2A = 0;
	OCR2B = 0;

	TCCR2A = 0xA3;	// Mode Fast PWM et Non-inverting 
	TCCR2B |= 0x02; // Frequence à 8kHz
}

//Interrupt sur pin A4 (PC4/PCINT12)
void init_interrupt()
{
	//PCINT1 pour echo ultrason
	PCICR |= _BV(PCIE1);	
	PCMSK1|= _BV(PCINT12);
	//INTO pour bouton poussoir
	EICRA|=0x03;
	EIMSK|= _BV(INT0);
}


void init_hardware()
{
	/*Sorties:
	PD4 (pin 4): AIN2
	PD5 (pin 5): AIN1
	PD6 (pin 6): cod
	PD7 (pin 7): cod
	PB0 (pin 8): LEDIR
	PB1 (pin 9): BIN1
	PB2 (pin10): BIN2
	PB4 (pin12): LEDLOW
	PB5 (pin13): LEDHIGH
	PC3 (pinA3): trig
	PB3/OC2A(pin11): PWMA : motG
	PD3/OC2B(pin 3): PWMB : motD
	
	Entrées:
	PD2/INTO: Interrupt BP	
	PCO : Phot0
	PC1 : Phot1
	PC2 : Phot2
	*/

	DDRB|=0x3F;
	DDRC|=0x08;
	DDRD|=0xF8;
	//Pull-up pour INTO
	PORTD|=0x04;

}

void init_timer()
{
	TCCR1B |= _BV(WGM12); 
	TCCR1B |= _BV(CS11);  // CS12 = 0; CS11 = 1; CS10 =1
	TCCR1B |= _BV(CS10);  // 250kHz
 	OCR1A   = NB_TICK;
	TIMSK1 |= _BV(OCIE1A);
}



void init_serial(void)
{
	uint8_t baudrate = BAUDRATE;
	UBRR0H = 0;
	UBRR0L = baudrate;
	UCSR0B = (1<<TXEN0);
	UCSR0C = 0x06;
}

void init_ADC()   
{   
	ADMUX|=_BV(REFS0)|_BV(ADLAR);

	ADCSRA|=_BV(ADIE)|  _BV(ADPS2) | _BV(ADPS1) | _BV(ADPS0);	//Conv complete interrupt enabled, ADC clock to 250kHz
 
	ADCSRA|= _BV(ADEN) | _BV(ADSC);	//Enable & start first conv
}  


//Commandes PWM des moteurs
void cmd_mot(int16_t cmdG,int16_t cmdD)
{

	if(cmdG<0)
	{
		PORTD|=0x10;	//AIN2 <- HIGH
		PORTD&=0XDF; 	//AIN1 <- LOW
		OCR2A = -cmdG;
	}
	else
	{
		PORTD&=0xEF;	//AIN2 <- LOW
		PORTD|=0X20; 	//AIN1 <- HIGH
		OCR2A = cmdG;
	}
	if(cmdD<0)
	{
		PORTB|=0x04;	//BIN2 <- HIGH
		PORTB&=0XFD; 	//BIN1 <- LOW
		OCR2B = -cmdD;
	}
	else
	{
		PORTB&=0xFB;	//BIN2 <- LOW
		PORTB|=0X02; 	//BIN1 <- HIGH
		OCR2B = cmdD;
	}
}


void send_serial(unsigned char c)
{
		loop_until_bit_is_set(UCSR0A, UDRE0);
		UDR0 = c;
}

void LEDlow(bool state)
{
	if(state)
		PORTB|=0x10;
	else
		PORTB&=0xEF;

}

void LEDhigh(bool state)
{
	if(state)
		PORTB|=0x20;
	else
		PORTB&=0xDF;
}	


//Conversions ADC en permanence
ISR(ADC_vect)
{
	switch (ADCchannel)
	{
		case 0:
			IR[0] = 0xFF - ADCH;
			ADMUX |= 0x01; 
			ADCchannel = 1;
			ADCSRA|= _BV(ADSC);
		break;

		case 1:
			IR[1] = 0xFF-ADCH;
			ADMUX &= 0xF0;	
			ADMUX |= 0x02;		
			ADCchannel = 2;	
			ADCSRA|= _BV(ADSC);
		break;

		case 2:
			IR[2] = 0xFF-ADCH;
			ADMUX &= 0xF0;
			ADCchannel = 0;
			ADCSRA|= _BV(ADSC);
		break;
	}
}


ISR(INT0_vect)
{
	ID++;
	ID%=3;
	ID+=1;
	//Affichage sur pins 12(PB4)LOW et 13 (PB5)HIGH
	switch(ID)
	{
		case 1:
			LEDlow(true);
			LEDhigh(false);
			IRperiodDetect=IRperiod0;
			IRperiodEmmit=IRperiod1;
			break;
		case 2:
			LEDlow(false);
			LEDhigh(true);
			IRperiodEmmit=IRperiod2;
			IRperiodDetect=IRperiod1;
			break;
		case 3:
			LEDlow(true);
			LEDhigh(true);
			IRperiodEmmit=IRperiod3;
			IRperiodDetect=IRperiod2;
			break;
	}
	_delay_ms(10);
}


//Interuption pour le echo de l'Ultrason
ISR(PCINT1_vect)
{
	if((PINC | 0xEF)==0xEF && UScount>0)
	{
		USdist= 1.5*UScount;
		UScount=-1000;
	}
}

//Intrruption Timer1 toutes les 100us
ISR(TIMER1_COMPA_vect)
{
	//Gestion Ultrason
	if (UScount==0)
		PORTC|=0x08;	//Envoi d'une onde ultrason
	if (UScount==1)
		PORTC&=0xF7;	//Fin d'envoi de l'onde ultrason
	UScount++;

	if(UScount==2000)
	{
		UScount=0;
		USdist=20;
	}

	//Gestion IR
	int i;
	//Pour chaque Phototransistor, on scrute les fronts montants
	for(i=0;i<3;i++)
	{
		IRcount[i]++;
		if (IR[i]>IRprec[i]+8) //Front montant, les variables IR[i] sont automatiquement mises à jour grâce à ISR(ACD_vect)
		{
			if(!IRcounting[i])	//Si on n'est pas en train de compter, on commence à compter
				{IRcounting[i]=true;IRcount[i]=0;}

			//Si un front montant arrive au bon moment
			if(IRcounting[i] && IRcount[i]>=(IRperiodDetect-10) && IRcount[i]<=(IRperiodDetect+10))
			{
				IRcount[i]=0;	//On réinitialise le compteur
				IRcollision[i]=false;
				// On considère qu'il faut 2 fronts montants au bon moment pour valider la détection du signal IR
				if (IRdetectedOnce[i])
					IRdetected[i]=true;
				else
					IRdetectedOnce[i]=true;
			}
		}

		//Si on n'a pas eu de front montant après la periode de détection, alors le signal IR est absent
		if(IRcounting[i] && IRcount[i]>=(IRperiodDetect+5))
		{	
			// On attend 2 erreurs avant de dire que le signal est perdu
			if(IRcollision[i])
			{
				IRcounting[i]=false;
				IRdetectedOnce[i]=false;
				IRdetected[i]=false;
			}
			else
			{IRcollision[i]=true;IRcount[i]=5;}
		}
		IRprec[i] = IR[i];
	}

	//Emission LEDIR: PB0
	IRcountEmmit++;
	if (IRcountEmmit==9)
		PORTB&=0xFE;
	else if (IRcountEmmit>=IRperiodEmmit)
		{PORTB|=0x01;IRcountEmmit=-1;}
}

void calculCommandes()
{
	int16_t cmdG=0,cmdD=0;
	int16_t cmd = (Kp*(USdist-CmdDist));
	if(cmd<-100){cmd=-100;}else if(cmd>100){cmd=100;}

	if(IRdetected[gauche] && !IRdetected[droit])
	{
		lastPos=gauche;
		if(IRdetected[milieu])
		{
			if(cmd<0){
			cmdD=-halfSpeed + cmd;
			cmdG=halfSpeed + cmd;}
			else{
			cmdD=halfSpeed + cmd;
			cmdG=-halfSpeed + cmd;}
		}
		else
		{
			cmdD=fullSpeed;
			cmdG=-fullSpeed;
		}
	}
	else if (!IRdetected[gauche] && IRdetected[droit])
	{
		lastPos=droit;
		if(IRdetected[milieu])
		{
			if(cmd<0){
			cmdD=halfSpeed + cmd;
			cmdG=-halfSpeed + cmd;}
			else{
			cmdD=-halfSpeed + cmd;
			cmdG=halfSpeed + cmd;}
		}
		else
		{
			cmdD=-fullSpeed;
			cmdG=fullSpeed;
		}
	}
	else if (!IRdetected[gauche] && !IRdetected[droit] && !IRdetected[milieu])
	{
		if (lastPos==droit)
		{
			cmdG=fullSpeed;
			cmdD=-fullSpeed;
		}
		else
		{
			cmdG=-fullSpeed;
			cmdD=fullSpeed;
		}

	}
	else if ((IRdetected[gauche] && IRdetected[droit]) || IRdetected[milieu])
	{
		cmdD=cmd;
		cmdG=cmd;
	}
	
	cmd_mot(cmdG,cmdD);
}

int main(void)
{
	init_hardware();
	init_timer();
	init_pwm();
	init_serial();
	init_interrupt();
	sei();
	init_ADC();
	while(1)
	{
		calculCommandes();
		_delay_ms(50);
	}
	cli();
	return 0;
}
